﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Globalization;

namespace GANNTuning.Data
{
    public class FakeType
    {
        private CSVReader.CSVRFake _fake = CSVReader.CSVRFake.Pulse;

        public FakeType(CSVReader.CSVRFake fk)
        {
            _fake = fk;
        }
        public CSVReader.CSVRFake Fake
        {
            get
            {
                return _fake;
            }
        }
        public override string ToString()
        {
            return Enum.GetName(typeof(CSVReader.CSVRFake), _fake);
        }
    }
    public class CSVReader
    {
        public enum CSVRFake
        {
            Random,
            Pulse,
            RandPulse,
            Equal,
            Noise
        }
        private Dictionary<int, List<double[]>> _patterns = null;
        private Dictionary<int, double> _results = null;
        private int _datainputs = 0;
        private double[][] _inputs = null;
        private double[][] _outputs = null;
        private int[] _values = null;

        public CSVReader(string filename)
        {
            StreamReader rdr = null;
            try
            {
                _patterns = new Dictionary<int, List<double[]>>();
                _results = new Dictionary<int, double>();
                rdr = new StreamReader(filename);
                List<string> ldata = new List<string>();
                string data;
                while ((data = rdr.ReadLine()) != null)
                {
                    ldata.Add(data);
                }
                int l = 0;
                string ds = CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator;
                for (int ix = 0; ix < ldata.Count; ix++)
                {
                    string[] values = ldata[ix].Split(ldata[ix].Contains(";") ? ';' : ',');
                    if (l <= 0)
                    {
                        l = values.Length;
                        _datainputs = l - 1;
                    }                    
                    if (l != values.Length)
                    {
                        throw new Exception(string.Format("Wrong number of values in line {0}", string.Join(";", values)));
                    }
                    int resp = int.Parse(values[l - 1]);
                    if (!_patterns.ContainsKey(resp))
                    {
                        _patterns[resp] = new List<double[]>();
                    }
                    double[] dvalues = new double[l];
                    for (int n = 0; n < l - 1; n++)
                    {
                        dvalues[n] = double.Parse(values[n].Replace(",", ds).Replace(".", ds));
                    }
                    _patterns[resp].Add(dvalues);
                }
                for (int sc = 0; sc < _datainputs; sc++)
                {
                    // Here the data is scaled, column by column
                    ScaleRng(sc, -1, 1);
                }
                // The output range is equally distributed among the patterns
                double dr = (0.99 + 0.99) / (_patterns.Keys.Count - 1);
                double response = 0.99;
                foreach (int r in _patterns.Keys)
                {
                    foreach (double[] s in _patterns[r])
                    {
                        s[s.Length - 1] = response;
                    }
                    _results[r] = response;
                    response -= dr;
                }
            }
            finally
            {
                if (rdr != null)
                {
                    rdr.Close();
                }
            }
        }
        public CSVReader(CSVRFake fake, int np, int nsamp, int ndata)
        {
            _patterns = new Dictionary<int, List<double[]>>();
            _results = new Dictionary<int, double>();
            _datainputs = ndata;
            double dr = (0.99 + 0.99) / (np - 1);
            double response = 0.99;
            double[] samp = null;
            Random r = new Random();
            switch (fake)
            {
                case CSVRFake.Pulse:
                    for (int pattern = 0; pattern < np; pattern++)
                    {
                        List<double[]> ld = new List<double[]>();
                        int nn = ndata / np;
                        for (int ns = 0; ns < nsamp; ns++)
                        {
                            samp = new double[ndata + 1];
                            double v = r.NextDouble();
                            for (int ix = 0; ix < ndata; ix++)
                            {                                
                                if ((ix >= (pattern * nn)) && (ix < (pattern + 1) * nn))
                                {
                                    samp[ix] = -v;
                                }
                                else
                                {
                                    samp[ix] = v;
                                }
                            }
                            samp[ndata] = response;
                            ld.Add(samp);
                        }
                        _patterns[pattern] = ld;
                        _results[pattern] = response;
                        response -= dr;
                    }
                    break;
                case CSVRFake.RandPulse:
                    for (int pattern = 0; pattern < np; pattern++)
                    {
                        List<double[]> ld = new List<double[]>();
                        int nn = ndata / np;
                        for (int ns = 0; ns < nsamp; ns++)
                        {
                            samp = new double[ndata + 1];
                            for (int ix = 0; ix < ndata; ix++)
                            {
                                double v = r.NextDouble();
                                if ((ix >= (pattern * nn)) && (ix < (pattern + 1) * nn))
                                {
                                    samp[ix] = -v;
                                }
                                else
                                {
                                    samp[ix] = v;
                                }
                            }
                            samp[ndata] = response;
                            ld.Add(samp);
                        }
                        _patterns[pattern] = ld;
                        _results[pattern] = response;
                        response -= dr;
                    }
                    break;
                case CSVRFake.Random:
                    for (int pattern = 0; pattern < np; pattern++)
                    {
                        List<double[]> ld = new List<double[]>();
                        for (int ns = 0; ns < nsamp; ns++)
                        {
                            samp = new double[ndata + 1];
                            for (int ix = 0; ix < ndata; ix++)
                            {
                                samp[ix] = (2 * r.NextDouble()) - 1;
                            }
                            samp[ndata] = response;
                            ld.Add(samp);
                        }
                        _patterns[pattern] = ld;
                        _results[pattern] = response;
                        response -= dr;
                    }
                    break;
                case CSVRFake.Equal:
                    List<double[]> ldc = new List<double[]>();
                    for (int ns = 0; ns < nsamp; ns++)
                    {
                        samp = new double[ndata + 1];
                        for (int ix = 0; ix < ndata; ix++)
                        {
                            samp[ix] = (2 * r.NextDouble()) - 1;
                        }
                        ldc.Add(samp);
                    }
                    for (int pattern = 0; pattern < np; pattern++)
                    {
                        List<double[]> ld = new List<double[]>();
                        for (int ns = 0; ns < nsamp; ns++)
                        {
                            samp = new double[ndata + 1];
                            Array.Copy(ldc[ns], samp, samp.Length);
                            samp[ndata] = response;
                            ld.Add(samp);
                        }
                        _patterns[pattern] = ld;
                        _results[pattern] = response;
                        response -= dr;
                    }
                    break;
                case CSVRFake.Noise:
                    for (int pattern = 0; pattern < np; pattern++)
                    {
                        List<double[]> ld = new List<double[]>();
                        int nn = ndata / np;
                        for (int ns = 0; ns < nsamp; ns++)
                        {
                            samp = new double[ndata + 1];
                            double v = r.NextDouble();
                            for (int ix = 0; ix < ndata; ix++)
                            {
                                double p = r.NextDouble();
                                if (p > 0.4)
                                {
                                    if ((ix >= (pattern * nn)) && (ix < (pattern + 1) * nn))
                                    {
                                        samp[ix] = -v;
                                    }
                                    else
                                    {
                                        samp[ix] = v;
                                    }
                                }
                                else
                                {
                                    samp[ix] = (2 * r.NextDouble()) - 1;
                                }
                            }
                            samp[ndata] = response;
                            ld.Add(samp);
                        }
                        _patterns[pattern] = ld;
                        _results[pattern] = response;
                        response -= dr;
                    }
                    break;
                default:
                    throw new Exception("Undefined fake set");
            }
        }
        public int ValueCount
        {
            get
            {
                return _datainputs;
            }
        }
        public int[] Patterns
        {
            get
            {
                int[] p = new int[_patterns.Keys.Count];
                int ix = 0;
                foreach (int m in _patterns.Keys)
                {
                    p[ix++] = m;
                }
                return p;
            }
        }
        public int MinSampleCount
        {
            get
            {
                int m = int.MaxValue;
                foreach (int p in _patterns.Keys)
                {
                    m = Math.Min(_patterns[p].Count, m);
                }
                return m;
            }
        }
        public int PatternCount
        {
            get
            {
                return _patterns.Count;
            }
        }
        public double MaxError
        {
            get
            {
                return 1.98 / _patterns.Count;
            }
        }
        public double[][] Inputs
        {
            get
            {
                return _inputs;
            }
        }
        public double[][] Outputs
        {
            get
            {
                return _outputs;
            }
        }
        public int[] Values
        {
            get
            {
                return _values;
            }
        }
        private void ScaleNorm(int iv)
        {
            double me = 0;
            double me2 = 0;
            double l = 0;
            foreach (int p in _patterns.Keys)
            {
                l += _patterns[p].Count;
                foreach (double[] s in _patterns[p])
                {
                    me += s[iv];
                    me2 += s[iv] * s[iv];
                }
            }
            me /= l;
            me2 /= l;
            double sd = Math.Sqrt(me2 - (me * me));
            foreach (int p in _patterns.Keys)
            {
                foreach (double[] s in _patterns[p])
                {
                    s[iv] = (s[iv] - me) / sd;
                }
            }
        }
        private void ScaleRng(int iv, double rmin, double rmax)
        {
            double max = double.MinValue;
            double min = double.MaxValue;
            foreach (int p in _patterns.Keys)
            {
                foreach (double[] s in _patterns[p])
                {
                    max = Math.Max(s[iv], max);
                    min = Math.Min(s[iv], min);
                }
            }
            double vrng = max - min;
            double rng = rmax - rmin;
            foreach (int p in _patterns.Keys)
            {
                foreach (double[] s in _patterns[p])
                {
                    s[iv] = (rng * (s[iv] - min) / vrng) + rmin;
                }
            }
        }
        public int PatternFromResult(double result)
        {
            double err = double.MaxValue;
            int pr = -1;
            foreach (int p in _results.Keys)
            {
                if (Math.Abs(result - _results[p]) < err)
                {
                    err = Math.Abs(result - _results[p]);
                    pr = p;
                }
            }
            return pr;
        }
        public void GetSample(int nm)
        {
            int cm = 0;
            foreach (int p in _patterns.Keys)
            {
                cm += Math.Min(_patterns[p].Count, nm);
            }
            Random r = new Random();
            _inputs = new double[cm][];
            _outputs = new double[cm][];
            _values = new int[cm];
            int b = 0;
            foreach (int p in _patterns.Keys)
            {
                List<double[]> samp = new List<double[]>();
                int cs = Math.Min(_patterns[p].Count, nm);
                for (int l = 0; l < cs; l++)
                {
                    _inputs[b] = new double[_datainputs];
                    _outputs[b] = new double[1];
                    int sm = r.Next(_patterns[p].Count);
                    double[] dm = _patterns[p][sm];
                    _patterns[p].RemoveAt(sm);
                    samp.Add(dm);
                    Array.Copy(dm, _inputs[b], _datainputs);
                    _outputs[b][0] = dm[dm.Length - 1];
                    _values[b++] = p;
                }
                _patterns[p].AddRange(samp);
            }
        }
    }
}
