﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading.Tasks;
using System.Windows.Forms;
using GANNTuning.Data;
using GANNTuning.Networks;
using GANNTuning.Genes;

namespace GANNTuning
{
    public partial class MainForm : Form
    {
        private const int _saveNN = 1;
        private const int _saveGENES = 2;
        private const int _saveGENEREPORT = 3;

        private Network _network = null;
        private CSVReader _data = null;
        private int _iterations = 0;
        private int _trials = 1;
        private double _learningrate = 0;
        private double _momentum = 0;
        private double _prevError = 0;
        private double _maxError = 0;
        private Bitmap _bmpErrors = null;
        private bool _break = false;
        private List<NNGene> _genes = null;

        public MainForm()
        {
            InitializeComponent();
            foreach (CSVReader.CSVRFake fk in Enum.GetValues(typeof(CSVReader.CSVRFake)))
            {
                cbFake.Items.Add(new FakeType(fk));
            }
            cbFake.SelectedIndex = 0;
        }
        private FileStream SaveFile(string filename, int index, string msg)
        {
            FileStream fs = null;
            BinaryFormatter fmt = new BinaryFormatter();
            switch (index)
            {
                case _saveNN:
                    fs = new FileStream(filename, FileMode.OpenOrCreate, FileAccess.ReadWrite);
                    if (_network != null)
                    {
                        fmt.Serialize(fs, _network);
                    }
                    break;
                case _saveGENES:
                    fs = new FileStream(filename, FileMode.OpenOrCreate, FileAccess.ReadWrite);
                    if (_genes != null)
                    {
                        _genes.Sort();
                        fmt.Serialize(fs, _genes);
                    }
                    break;
                case _saveGENEREPORT:
                    if (_genes != null)
                    {
                        _genes.Sort();
                        StreamWriter wr = new StreamWriter(filename);
                        wr.WriteLine("Alpha;Input;Hidden;Cost;Min Error;Mid Error;Max Error; Mean Error;SD Error");
                        for (int ix = 0; ix < _genes.Count; ix++)
                        {
                            wr.WriteLine(string.Join(";", new string[] { _genes[ix].Alpha.ToString(), _genes[ix].Input, _genes[ix].Hidden, _genes[ix].Cost.ToString(),
                                _genes[ix].MinErrNN.ToString(), _genes[ix].MidErrNN.ToString(), _genes[ix].MaxErrNN.ToString(), _genes[ix].MeanErrNN.ToString(), _genes[ix].DvErrNN.ToString()}));
                        }
                        wr.Close();
                    }
                    break;
            }
            if (!string.IsNullOrEmpty(msg))
            {
                MessageBox.Show(msg);
            }
            return fs;
        }

        private void InitializeFeedback()
        {
            if (cbTrack.Checked)
            {
                if (_bmpErrors == null)
                {
                    _bmpErrors = new Bitmap(pbErrors.Width, pbErrors.Height);
                }
                using (Graphics gr = Graphics.FromImage(_bmpErrors))
                {
                    gr.FillRectangle(Brushes.White, new Rectangle(0, 0, _bmpErrors.Width, _bmpErrors.Height));
                }
                _maxError = 2;
                _prevError = _maxError;
                pbProgress.Value = 0;
                pbProgress.Maximum = _iterations;
                lIterations.Text = "0";
                _network.ReportProgress += new EventHandler(TrainingFeedback);
            }
        }

        private Network BuildNetwork()
        {
            int ninput = 0;
            int[] nhidden = null;
            double alpha = 0;
            if ((!int.TryParse(tbInput.Text, out ninput)) || (ninput <= 0))
            {
                tcData.SelectedIndex = 0;
                tbInput.Focus();
                throw new Exception("Wrong number of input neurons");
            }
            if (!string.IsNullOrEmpty(tbHidden.Text))
            {
                string[] shidden = tbHidden.Text.Split(',');
                nhidden = new int[shidden.Length];
                for (int h = 0; h < shidden.Length; h++)
                {
                    if ((!int.TryParse(shidden[h].Trim(), out nhidden[h])) || (nhidden[h] <= 0))
                    {
                        tcData.SelectedIndex = 0;
                        tbHidden.Focus();
                        throw new Exception("Wrong number of hidden neurons");
                    }
                }
            }
            if (!double.TryParse(tbAlpha.Text, out alpha))
            {
                tcData.SelectedIndex = 0;
                tbAlpha.Focus();
                throw new Exception("Wrong alpha value");
            }
            alpha = Math.Max(0.001, Math.Min(50, alpha));
            tbAlpha.Text = alpha.ToString();
            Network n = null;
            if (_data != null)
            {
                n = new Network(_data.ValueCount, ninput, nhidden, alpha);
            }
            else
            {
                throw new Exception("You must load a data set first");
            }
            return n;
        }

        private int GetTrainParams()
        {
            _iterations = 1000;
            if (!int.TryParse(tbIterations.Text, out _iterations))
            {
                tcData.SelectedIndex = 2;
                tbIterations.Focus();
                throw new Exception("Wrong number of iterations");
            }
            _iterations = Math.Max(100, _iterations);
            tbIterations.Text = _iterations.ToString();
            _learningrate = 0;
            if (!double.TryParse(tbLR.Text, out _learningrate))
            {
                tcData.SelectedIndex = 2;
                tbLR.Focus();
                throw new Exception("Wrong learning rate");
            }
            _learningrate = Math.Max(0.00001, Math.Min(1, _learningrate));
            tbLR.Text = _learningrate.ToString();
            _momentum = 0;
            if (!double.TryParse(tbMomentum.Text, out _momentum))
            {
                tcData.SelectedIndex = 2;
                tbMomentum.Focus();
                throw new Exception("Wrong momentum");
            }
            _momentum = Math.Max(0, Math.Min(0.5, _momentum));
            tbMomentum.Text = _momentum.ToString();
            int pc = 70;
            if (int.TryParse(tbPD.Text, out pc))
            {
                pc = Math.Min(90, Math.Max(10, pc));
            }
            tbPD.Text = pc.ToString();
            return pc;
        }

        private async Task TrainNetwork()
        {
            int pc = GetTrainParams();
            if (_data != null)
            {
                int sc = (pc * _data.MinSampleCount) / 100;
                _data.GetSample(sc);
                InitializeFeedback();
                double error = await _network.TrainNetwork(_data.Inputs, _data.Outputs, _learningrate, _momentum, _iterations);
                _network.ComputeErrors(_data.Inputs, _data.Outputs, _data.MaxError);
            }
            else
            {
                throw new Exception("No data loaded");
            }
        }
        private void EnableControls(bool enable)
        {
            bNew.Enabled = enable;
            bOpen.Enabled = enable;
            bSave.Enabled = enable;
            bDeleteNN.Enabled = enable;
            bTrain.Enabled = enable;
            bCompute.Enabled = enable;
            bGenes.Enabled = enable;            
            bStopTraining.Enabled = !enable;
        }

        private async Task<Network> TrainGene(NNGene gen, int trials)
        {
            Network nn = null;
            _network = gen.BuildNN(_data.ValueCount);
            for (int fc = 0; fc < trials; fc++)
            {
                _network.InitializeWeights();
                InitializeFeedback();
                await _network.TrainNetwork(_data.Inputs, _data.Outputs, _learningrate, _momentum, _iterations);
                _network.ComputeErrors(_data.Inputs, _data.Outputs, _data.MaxError);                
                if (gen.ComputeCost(_network))
                {
                    nn = _network;
                }
            }
            return nn;
        }
        private void TrainingFeedback(object sender, EventArgs e)
        {
            if (_network.TrainIteration > 0)
            {
                float ix = _bmpErrors.Width / (float)_iterations;
                float iy = _bmpErrors.Height / (float)_maxError;
                using (Graphics gr = Graphics.FromImage(_bmpErrors))
                {
                    gr.DrawLine(Pens.Black,
                        new PointF((_network.TrainIteration - 1f) * ix, _bmpErrors.Height - (float)_prevError * iy),
                        new PointF(_network.TrainIteration * ix, _bmpErrors.Height - (float)_network.TrainError * iy));
                }
                _prevError = _network.TrainError;
                pbProgress.Invoke((Action)(() => { pbProgress.Value = _network.TrainIteration + 1; }));
                lIterations.Text = (_network.TrainIteration + 1).ToString() + "     Err: " + _network.TrainError.ToString();                
                pbErrors.Invoke((Action)(() => { pbErrors.Image = _bmpErrors.Clone() as Bitmap; }));
            }            
        }

        private void bNew_Click(object sender, EventArgs e)
        {
            try
            {
                int np = 0;
                if (!int.TryParse(tbNpats.Text, out np))
                {
                    tcData.SelectedIndex = 2;
                    tbNpats.Focus();
                    throw new Exception("Wrong number of patterns");
                }
                int nsamp = 0;
                if (!int.TryParse(tbSamples.Text, out nsamp))
                {
                    tcData.SelectedIndex = 2;
                    tbSamples.Focus();
                    throw new Exception("Wrong number of samples");
                }
                int ndata = 0;
                if (!int.TryParse(tbNValues.Text, out ndata))
                {
                    tcData.SelectedIndex = 2;
                    tbNValues.Focus();
                    throw new Exception("Wrong number of values");
                }
                int nv = 0;
                if (_data != null)
                {
                    nv = _data.ValueCount;
                }
                _data = new CSVReader(((FakeType)cbFake.SelectedItem).Fake, np, nsamp, ndata);
                if (nv != _data.ValueCount)
                {
                    _network = null;
                }
                MessageBox.Show("Data created");
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void bOpen_Click(object sender, EventArgs e)
        {
            try
            {
                if (ofDlg.ShowDialog() == DialogResult.OK)
                {
                    if (Path.GetExtension(ofDlg.FileName).ToLower().EndsWith("csv"))
                    {
                        int nv = 0;
                        if (_data != null)
                        {
                            nv = _data.ValueCount;
                        }
                        _data = new CSVReader(ofDlg.FileName);
                        MessageBox.Show("Data loaded");
                        if (nv != _data.ValueCount)
                        {
                            _network = null;
                        }
                        return;
                    }
                    FileStream fs = null;
                    try
                    {
                        BinaryFormatter fmt = new BinaryFormatter();
                        fs = new FileStream(ofDlg.FileName, FileMode.Open, FileAccess.ReadWrite);
                        if (Path.GetExtension(ofDlg.FileName).ToLower().EndsWith("net"))
                        {
                            _network = fmt.Deserialize(fs) as Network;
                            string[] np = _network.Params.Split(';');
                            tbInput.Text = np[0];
                            tbHidden.Text = np[1];
                            tbAlpha.Text = np[2];
                            MessageBox.Show("Network loaded");
                            return;
                        }
                        if (Path.GetExtension(ofDlg.FileName).ToLower().EndsWith("gen"))
                        {
                            _genes = fmt.Deserialize(fs) as List<NNGene>;
                            MessageBox.Show("Genes loaded");
                            return;
                        }
                    }
                    catch
                    {
                        throw new Exception("Unknown file format");
                    }
                    finally
                    {
                        if (fs != null)
                        {
                            fs.Close();
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void bSave_Click(object sender, EventArgs e)
        {
            FileStream fs = null;
            try
            {
                if (sfDlg.ShowDialog() == DialogResult.OK)
                {
                    fs = SaveFile(sfDlg.FileName, sfDlg.FilterIndex, sfDlg.FilterIndex == 1 ? "Network saved" : "Genes saved");
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                if (fs != null)
                {
                    fs.Close();
                }
            }
        }
        private void bDeleteNN_Click(object sender, EventArgs e)
        {
            _network = null;
            MessageBox.Show("Network Deleted");
        }
        private async void bTrain_Click(object sender, EventArgs e)
        {
            try
            {
                EnableControls(false);
                tcData.SelectedIndex = 3;
                if (_network == null) 
                {
                    _network = BuildNetwork();
                }
                else
                {
                    string[] p = _network.Params.Split(';');
                    if ((p[0] != tbInput.Text) ||
                        (p[1] != tbHidden.Text) ||
                        (p[2] != tbAlpha.Text))
                    {
                        _network = BuildNetwork();
                    }
                }
                await TrainNetwork();
                MessageBox.Show("Network trained");
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                EnableControls(true);
            }
        }

        private void bStopTraining_Click(object sender, EventArgs e)
        {
            try
            {
                _network.Break();
                _break = true;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void bCompute_Click(object sender, EventArgs e)
        {
            try
            {
                pResults.Height = 200;
                if (_network == null)
                {
                    throw new Exception("No network present");
                }
                if (_data == null)
                {
                    throw new Exception("No data present");
                }
                tcData.SelectedIndex = 4;
                Bitmap bmp = new Bitmap(pbGraph.Width, pbGraph.Height);
                GetTrainParams();
                List<double> errs = new List<double>();
                _data.GetSample(_data.MinSampleCount);
                string info = string.Format("Samples: {0}", _data.Outputs.Length.ToString());
                double[][] results = _network.ComputeErrors(_data.Inputs, _data.Outputs, _data.MaxError);
                int ok = 0;
                int bad = 0;
                int[,] ptcnt = new int[1 + _data.PatternCount, 1 + _data.PatternCount];
                int cp = 1;
                foreach (int p in _data.Patterns)
                {
                    ptcnt[cp, 0] = p;
                    ptcnt[0, cp] = p;
                    cp++;
                }
                for (int ie = 0; ie < results.Length; ie++)
                {
                    errs.Add(Math.Abs(_data.Outputs[ie][0] - results[ie][0]));
                    int vr = _data.PatternFromResult(results[ie][0]);
                    if (vr == _data.Values[ie])
                    {
                        ok++;
                    }
                    else
                    {
                        bad++;
                    }
                    for (int pp = 1; pp <= ptcnt.GetUpperBound(0); pp++)
                    {
                        if (ptcnt[pp, 0] == _data.Values[ie])
                        {
                            for (int pr = 1; pr <= ptcnt.GetUpperBound(1); pr++)
                            {
                                if (ptcnt[0, pr] == vr)
                                {
                                    ptcnt[pp, pr]++;
                                    break;
                                }
                            }
                            break;
                        }
                    }
                }
                info += "\nOK: " + ok.ToString() + " / NOK: " + bad.ToString() + "  (" + ((100 * bad) / (float)results.Length).ToString("0.##") + "% Error)\n\n";
                info += "Detail:\n        ";
                for (int pr = 1; pr <= ptcnt.GetUpperBound(1); pr++)
                {
                    info += ptcnt[0, pr].ToString().PadLeft(8);
                }
                info += "\n";
                for (int pp = 1; pp <= ptcnt.GetUpperBound(0); pp++)
                {
                    for (int pr = 0; pr <= ptcnt.GetUpperBound(1); pr++)
                    {
                        info += ptcnt[pp, pr].ToString().PadLeft(8);
                    }
                    info += "\n";
                }
                NNGene costgene = new NNGene(_network, true);
                info += "\nCost      : " + costgene.Cost.ToString();
                info += "\nMin Error : " + _network.TrainErrorMin.ToString();
                info += "\nMid Error : " + _network.TrainErrorMedian.ToString();
                info += "\nMean Error: " + _network.TrainError.ToString();
                info += "\nMax Error : " + _network.TrainErrorMax.ToString();
                info += "\nSD Error  : " + _network.TrainErrorDev.ToString();
                tbResults.Lines = info.Split('\n');
                errs.Sort();
                double de = errs[errs.Count - 1] / 100;
                int[] lcnt = new int[100];
                int ixfatalerr = (int)Math.Truncate((1.98 / _data.PatternCount) / de);
                for (int ie = 0; ie < errs.Count; ie++)
                {
                    lcnt[Math.Min(99, (int)Math.Truncate(errs[ie] / de))]++;
                }
                using (Graphics gr = Graphics.FromImage(bmp))
                {
                    gr.FillRectangle(Brushes.White, new Rectangle(0, 0, bmp.Width, bmp.Height));
                    float dy = (float)bmp.Height / lcnt.Max();
                    float dx = bmp.Width / 100f;
                    for (int x = 0; x < 100; x++)
                    {
                        Brush br = Brushes.Blue;
                        if (x >= ixfatalerr)
                        {
                            br = Brushes.Red;
                        }
                        gr.FillRectangle(br, x * dx, pbGraph.Height - (lcnt[x] * dy), dx, lcnt[x] * dy);
                    }
                }
                pbGraph.Image = bmp;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private async void bGenes_Click(object sender, EventArgs e)
        {
            try
            {
                Network bestnn = _network;
                NNGene bestgene = null;
                if (_network != null)
                {
                    bestgene = new NNGene(_network, true);
                }
                else
                {
                    bestgene = new NNGene();
                }
                pResults.Height = 52;
                _bmpErrors = null;
                EnableControls(false);
                tcData.SelectedIndex = 3;
                Bitmap bmp = new Bitmap(pbGraph.Width, pbGraph.Height);
                int pop = 0;
                if (!int.TryParse(tbPopulation.Text, out pop))
                {
                    tcData.SelectedIndex = 1;
                    tbPopulation.Focus();
                    throw new Exception("Bad population");
                }
                pop = Math.Min(100, pop);
                tbPopulation.Text = pop.ToString();
                int maxnn = 0;
                if (!int.TryParse(tbMaxNN.Text, out maxnn))
                {
                    tcData.SelectedIndex = 1;
                    tbMaxNN.Focus();
                    throw new Exception("Bad max. neurons");
                }
                int maxh = 0;
                if (!int.TryParse(tbMaxHidden.Text, out maxh))
                {
                    tcData.SelectedIndex = 1;
                    tbMaxHidden.Focus();
                    throw new Exception("Bad max. hidden");
                }
                _trials = 1;
                if (!int.TryParse(tbTrials.Text, out _trials))
                {
                    tcData.SelectedIndex = 1;
                    tbTrials.Focus();
                    throw new Exception("Bad trial count");
                }
                int pc = GetTrainParams();
                _data.GetSample((pc * _data.MinSampleCount) / 100);

                double mincost = double.MaxValue;
                double maxcost = double.MinValue;
                NNGene gene = null;                

                if (_genes == null)
                {
                    _genes = new List<NNGene>();
                    // Create genes randomly
                    for (int i = 0; i < pop; i++)
                    {
                        gene = new NNGene(null, null, maxnn, maxh);
                        _genes.Add(gene);
                        Network nn = await TrainGene(gene, _trials);
                        if ((nn != null) && (bestgene.Cost > gene.Cost))
                        {
                            bestnn = nn;
                            bestgene = new NNGene(nn, true);
                        }
                        mincost = Math.Min(mincost, gene.Cost);
                        maxcost = Math.Max(maxcost, gene.Cost);
                        tbResults.Text = "Population: " + _genes.Count.ToString();
                    }
                }
                else
                {
                    bestgene = _genes[0];
                    foreach (NNGene g in _genes)
                    {
                        mincost = Math.Min(mincost, g.Cost);
                        maxcost = Math.Max(maxcost, g.Cost);
                    }
                }
                
                List<double> errs = new List<double>();
                _break = false;
                int generation = 0;
                List<NNGene> nwgen = new List<NNGene>();
                while (!_break)
                {
                    _genes.Sort();
                    generation++;
                    if (errs.Count >= pbGraph.Width)
                    {
                        errs.Clear();
                    }
                    errs.Add(_genes[0].Cost);
                    tbInput.Text = _genes[0].Input;
                    tbHidden.Text = _genes[0].Hidden;
                    tbAlpha.Text = _genes[0].Alpha;                    
                    double step = pbGraph.Height / (maxcost - mincost);
                    using (Graphics gr = Graphics.FromImage(bmp))
                    {
                        gr.FillRectangle(Brushes.White, new Rectangle(0, 0, pbGraph.Width, pbGraph.Height));
                        for (int x = 0; x < errs.Count; x++)
                        {
                            Pen p = Pens.Red;
                            gr.DrawLine(p, x, (float)(pbGraph.Height - (errs[x] * step)), x, pbGraph.Height);
                        }
                        pbGraph.Invoke((Action)(() => { pbGraph.Image = bmp; }));
                    }
                    nwgen.Clear();
                    // Cross the three best genes each other
                    for (int g = 0; g < 3; g++)
                    {
                        for (int ix = 1; ix <= 5; ix++)
                        {
                            gene = new NNGene(_genes[g], _genes[g + ix], maxnn, maxh);
                            nwgen.Add(gene);
                            Network nn = await TrainGene(gene, _trials);
                            if ((nn != null) && (bestgene.Cost > gene.Cost))
                            {
                                bestnn = nn;
                                bestgene = new NNGene(nn, true);
                            }
                            mincost = Math.Min(mincost, gene.Cost);
                            maxcost = Math.Max(maxcost, gene.Cost);
                            tbResults.Text = "Min Cost: " + mincost.ToString() + " / Max Cost: " + maxcost.ToString() + " / Gen.: " + generation.ToString() + " / New Pop.: " + nwgen.Count.ToString();
                        }
                    }
                    // Cross the three best genes randomly
                    Random rc = new Random();
                    for (int g = 0; g < 3; g++)
                    {
                        for (int ir = 0; ir < 10; ir++)
                        {
                            int ng = rc.Next(g + 1, _genes.Count);
                            gene = new NNGene(_genes[g], _genes[ng], maxnn, maxh);
                            nwgen.Add(gene);
                            Network nn = await TrainGene(gene, _trials);
                            if ((nn != null) && (bestgene.Cost > gene.Cost))
                            {
                                bestnn = nn;
                                bestgene = new NNGene(nn, true);
                            }
                            mincost = Math.Min(mincost, gene.Cost);
                            maxcost = Math.Max(maxcost, gene.Cost);
                            tbResults.Text = "Min Cost: " + mincost.ToString() + " / Max Cost: " + maxcost.ToString() + " / Gen.: " + generation.ToString() + " / New Pop.: " + nwgen.Count.ToString();
                        }
                    }
                    // Cross 25% of genes randomly
                    int np = (25 * pop) / 100;
                    for (int ix = 0; ix < np; ix++)
                    {
                        gene = new NNGene(_genes[rc.Next(_genes.Count)], _genes[rc.Next(_genes.Count)], maxnn, maxh);
                        nwgen.Add(gene);
                        Network nn = await TrainGene(gene, _trials);
                        if ((nn != null) && (bestgene.Cost > gene.Cost))
                        {
                            bestnn = nn;
                            bestgene = new NNGene(nn, true);
                        }
                        mincost = Math.Min(mincost, gene.Cost);
                        maxcost = Math.Max(maxcost, gene.Cost);
                        tbResults.Text = "Min Cost: " + mincost.ToString() + " / Max Cost: " + maxcost.ToString() + " / Gen.: " + generation.ToString() + " / New Pop.: " + nwgen.Count.ToString();
                    }
                    // Create new 25% genes randomly
                    np = (25 * pop) / 100;
                    for (int ix = 0; ix < np; ix++)
                    {
                        gene = new NNGene(null, null, maxnn, maxh);
                        nwgen.Add(gene);
                        Network nn = await TrainGene(gene, _trials);
                        if ((nn != null) && (bestgene.Cost > gene.Cost))
                        {
                            bestnn = nn;
                            bestgene = new NNGene(nn, true);
                        }
                        mincost = Math.Min(mincost, gene.Cost);
                        maxcost = Math.Max(maxcost, gene.Cost);
                        tbResults.Text = "Min Cost: " + mincost.ToString() + " / Max Cost: " + maxcost.ToString() + " / Gen.: " + generation.ToString() + " / New Pop.: " + nwgen.Count.ToString();
                    }
                    // Delete worse genes
                    int sup = Math.Max(0, (_genes.Count - nwgen.Count));
                    _genes.RemoveRange(sup, _genes.Count - sup);
                    // Give the rest a new opportunity to improve
                    for (int g = 0; g < _genes.Count; g++)
                    {
                        Network nn = await TrainGene(_genes[g], 1);
                        if ((nn != null) && (bestgene.Cost > gene.Cost))
                        {
                            bestnn = nn;
                            bestgene = new NNGene(nn, true);
                        }
                        tbResults.Text = "Min Cost: " + mincost.ToString() + " / Max Cost: " + maxcost.ToString() + " / Gen.: " + generation.ToString() + " / Pop.: " + _genes.Count.ToString() + " / Proc.: " + (g + 1).ToString();
                    }
                    // Add new genes
                    _genes.AddRange(nwgen);
                    // Save new generation
                    FileStream fs = null;
                    try
                    {
                        fs = SaveFile("gentmp.gen", _saveGENES, null);
                        _network = bestnn;
                    }
                    catch
                    {
                    }
                    finally
                    {
                        if (fs != null)
                        {
                            fs.Close();
                            fs = null;
                        }
                    }
                    // Save best network
                    try
                    {
                        fs = SaveFile("gentmp.net", _saveNN, null);
                    }
                    catch
                    {
                    }
                    finally
                    {
                        if (fs != null)
                        {
                            fs.Close();
                        }
                    }
                }                
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                EnableControls(true);
            }
        }
    }
}
