﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

namespace GANNTuning.Networks
{
    [Serializable]
    public class Network
    {
        private Layer[] _layers = null;
        private double _alpha = 2;
        private int _iteration = 0;
        private double _error = 0;
        private double _errDev = 0;
        private double _errmedian = 0;
        private double _errmin = 0;
        private double _errmax = 0;
        private double _perror = 0;
        private bool _run = true;

        public Network()
        {
        }
        public Network(int datainputs, int inputs, int[] hidden, double alpha)
        {
            _layers = new Layer[2 + ((hidden != null) ? hidden.Length : 0)];
            _layers[0] = new Layer(datainputs, inputs, alpha);
            Layer prev = _layers[0];
            if ((hidden != null) && (hidden.Length > 0))
            {
                for (int h = 0; h < hidden.Length; h++)
                {
                    _layers[1 + h] = new Layer(hidden[h], alpha, prev);
                    prev = _layers[1 + h];
                }
            }
            _layers[_layers.Length - 1] = new Layer(1, alpha, prev);
            _alpha = alpha;
        }        
        public event EventHandler ReportProgress = null;

        public int Neurons
        {
            get
            {
                int n = 0;
                for (int i = 0; i < _layers.Length; i++)
                {
                    n += _layers[i].OutputCount;
                }
                return n;
            }
        }
        public int Layers
        {
            get
            {
                return _layers.Length;
            }
        }
        public int Inputs
        {
            get
            {
                return _layers[0].OutputCount;
            }
        }
        public int[] Hidden
        {
            get
            {
                if (_layers.Length < 3)
                {
                    return null;
                }
                int[] hl = new int[_layers.Length - 2];
                for (int h = 1; h < _layers.Length - 1; h++)
                {
                    hl[h - 1] = _layers[h].OutputCount;
                }
                return hl;
            }
        }
        public double Alpha
        {
            get
            {
                return _alpha;
            }
        }
        public string Params
        {
            get
            {
                string p = _layers[0].OutputCount.ToString() + ";";
                string pc = _layers.Length > 2 ? "" : ";";
                string c = "";
                for (int ix = 1; ix < _layers.Length - 1; ix++)
                {
                    p += c + _layers[ix].OutputCount.ToString();
                    c = ",";
                    pc = ";";
                }
                p += pc + _alpha.ToString();
                return p;
            }
        }
        public int TrainIteration
        {
            get
            {
                return _iteration;
            }
        }
        public double TrainError
        {
            get
            {
                return _error;
            }
        }
        public double TrainErrorDev
        {
            get
            {
                return _errDev;
            }
        }
        public double TrainErrorMin
        {
            get
            {
                return _errmin;
            }
        }
        public double TrainErrorMedian
        {
            get
            {
                return _errmedian;
            }
        }
        public double TrainErrorMax
        {
            get
            {
                return _errmax;
            }
        }
        public double FailRate
        {
            get
            {
                return _perror;
            }
        }
        public void Break()
        {
            _run = false;
        }
        public void SetMomentum(double m)
        {
            for (int l = 0; l < _layers.Length; l++)
            {
                _layers[l].Momentum = m;
            }
        }
        public void InitializeWeights()
        {
            for (int l = 0; l < _layers.Length; l++)
            {
                _layers[l].InitializeWeights();
            }
        }
        private Task<double> TrainSample(double[] inputs, double[] finaloutputs, double learningrate)
        {
            return Task.Run(() =>
            {
                _layers[0].Process(inputs);
                double error = _layers[_layers.Length - 1].ComputeError(finaloutputs);
                _layers[0].ComputeUpdates(inputs, learningrate);
                _layers[0].ApplyUpdates();
                return error;
            });
        }
        public async Task<double> TrainNetwork(double[][] inputs, double[][] outputs, double learningrate, double momentum, int iterations)
        {
            try
            {
                // Randomize samples
                List<int> lindexes = new List<int>();
                Random r = new Random();
                for (int ix = 0; ix < inputs.Length; ix++)
                {
                    lindexes.Insert(r.Next(ix), ix);
                }                         
                _run = true;
                _error = 0;
                SetMomentum(momentum);
                for (_iteration = 0; _iteration < iterations; _iteration++)
                {                    
                    List<int> lindexes2 = new List<int>();
                    double error = 0;
                    for (int s = 0; s < inputs.Length; s++)
                    {
                        if (!_run)
                        {
                            return _error;
                        }
                        // Get an index randomly
                        int ix = r.Next(lindexes.Count);
                        // Save the used index in random order
                        int ixr = lindexes[ix];
                        lindexes2.Insert(r.Next(lindexes2.Count), ixr);                        
                        // Do not reuse index
                        lindexes.RemoveAt(ix);
                        error += await TrainSample(inputs[ixr], outputs[ixr], learningrate);
                    }
                    lindexes = lindexes2;
                    _error = error / inputs.Length;
                    ReportProgress?.Invoke(this, EventArgs.Empty);
                }
                return _error;
            }
            finally
            {
                ReportProgress = null;
            }
        }
        public double[][] ComputeErrors(double[][] inputs, double[][] outputs, double maxerror)
        {
            List<double> errors = new List<double>();
            _perror = 0;
            // Compute final errors
            double[][] responses = Compute(inputs);
            for (int s = 0; s < responses.Length; s++)
            {
                double cerr = Math.Abs(responses[s][0] - outputs[s][0]);
                errors.Add(cerr);
                // Count bad responses
                if (cerr >= maxerror)
                {
                    _perror += 1;
                }
            }
            // Calculate mean error
            _error = errors.Sum() / inputs.Length;
            // Calculate error variance
            double evar = 0;
            for (int ixd = 0; ixd < errors.Count; ixd++)
            {
                double v = errors[ixd] - _error;
                evar += v * v;
            }
            // Probability of a wrong response
            _perror /= errors.Count;
            // Errors standard deviation
            _errDev = Math.Sqrt(evar / errors.Count);
            errors.Sort();
            // Minimum error
            _errmin = errors[0];
            // Median error
            _errmedian = errors[errors.Count / 2];
            // Maximum error
            _errmax = errors[errors.Count - 1];
            return responses;
        }
        public double[][] Compute(double[][] inputs)
        {
            double[][] result = new double[inputs.Length][];
            for (int ix = 0; ix < inputs.Length; ix++)
            {
                _layers[0].Process(inputs[ix]);
                result[ix] = new double[_layers[_layers.Length - 1].Outputs.Length];
                _layers[_layers.Length - 1].Outputs.CopyTo(result[ix], 0);
            }
            return result;
        }
    }
}
