﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using GANNTuning.Networks;

namespace GANNTuning.Genes
{
    [Serializable]
    public class NNGene : IEquatable<NNGene>, IComparable<NNGene>
    {
        private double _fcost = double.MaxValue;
        private double _minerrNN = 0;
        private double _maxerrNN = 0;
        private double _meanerrNN = 0;
        private double _miderrNN = 0;
        private double _dverrNN = 0;
        private int _input = 0;
        private int[] _hidden = null;
        private double _alpha = 0;
        public NNGene()
        {
        }
        public NNGene(Network net, bool compute)
        {
            _input = net.Inputs;
            _hidden = net.Hidden;
            _alpha = net.Alpha;
            if (compute)
            {
                ComputeCost(net);
            }
        }
        public NNGene(NNGene g1, NNGene g2, int maxn, int maxh)
        {
            Random r = new Random();
            if ((g1 == null) || (g2 == null))
            {
                CreateNewGene(maxn, maxh);
            }
            else
            {
                double p = r.NextDouble();
                if (p < 0.5)
                {
                    Cross(g1, g2);
                }
                else
                {
                    Cross(g2, g1);
                }
                p = r.NextDouble();
                if ((p < 0.05) || SameAs(g1) || SameAs(g2))
                {
                    Mutate(maxn, maxh);
                }
            }
        }
        public double Cost
        {
            get
            {
                return _fcost;
            }
        }
        public double MinErrNN
        {
            get
            {
                return _minerrNN;
            }
            set
            {
                _minerrNN = value;
            }
        }
        public double MaxErrNN
        {
            get
            {
                return _maxerrNN;
            }
            set
            {
                _maxerrNN = value;
            }
        }
        public double MeanErrNN
        {
            get
            {
                return _meanerrNN;
            }
            set
            {
                _meanerrNN = value;
            }
        }
        public double MidErrNN
        {
            get
            {
                return _miderrNN;
            }
            set
            {
                _miderrNN = value;
            }
        }
        public double DvErrNN
        {
            get
            {
                return _dverrNN;
            }
            set
            {
                _dverrNN = value;
            }
        }
        public string Input
        {
            get
            {
                return _input.ToString();
            }
        }
        public string Hidden
        {
            get
            {
                string s = "";
                if (_hidden != null)
                {
                    string c = "";
                    for (int h = 0; h < _hidden.Length; h++)
                    {
                        s += c + _hidden[h].ToString();
                        c = ",";
                    }
                }
                return s;
            }
        }
        public string Alpha
        {
            get
            {
                return _alpha.ToString();
            }
        }
        private void CreateNewGene(int maxn, int maxh)
        {
            Random r = new Random();
            _input = r.Next(3, maxn + 1);
            int nh = r.Next(maxh + 1);
            if (nh > 0)
            {
                _hidden = new int[nh];
                for (int l = 0; l < nh; l++)
                {
                    _hidden[l] = r.Next(1, maxn + 1);
                }
            }
            double da = r.NextDouble();
            double sgn = r.NextDouble() > 0.5 ? 1 : -1;
            _alpha = 2 + (sgn * Math.Round(da * da * da, 2));
        }
        private void Cross(NNGene g1, NNGene g2)
        {
            _input = g1._input;
            if (g1._hidden == null)
            {
                _hidden = g2._hidden;
            }
            else
            {
                if (g2._hidden == null)
                {
                    _hidden = g1._hidden;
                }
                else
                {
                    _hidden = new int[Math.Min(g1._hidden.Length, g2._hidden.Length)];
                    for (int h = 0; h < _hidden.Length; h++)
                    {
                        if ((h & 1) != 0)
                        {
                            _hidden[h] = g2._hidden[h];
                        }
                        else
                        {
                            _hidden[h] = g1._hidden[h];
                        }
                    }
                }
            }
            _alpha = g2._alpha;
        }
        private void Mutate(int maxn, int maxh)
        {
            Random r = new Random();
            double da = r.NextDouble();
            double p = r.NextDouble();
            _alpha = Math.Max(0.5, _alpha + (p < 0.5 ? Math.Round(da * da * da, 2) : -Math.Round(da * da * da, 2)));
            p = r.NextDouble();
            if (p < 0.3)
            {
                _input = r.Next(3, maxn + 1);
            }
            p = r.NextDouble();
            if (p < 0.3)
            {
                int nh = r.Next(maxh + 1);
                if (nh > 0)
                {
                    _hidden = new int[nh];
                    for (int l = 0; l < nh; l++)
                    {
                        _hidden[l] = r.Next(1, maxn + 1);
                    }
                }
            }

        }
        public bool ComputeCost(Network net)
        {
            string[] pnet = net.Params.Split(';');
            if ((Input != pnet[0]) ||
                (Hidden != pnet[1]) ||
                (_alpha != net.Alpha))
            {
                throw new Exception("Network and gene params do not match");
            }
            double cost = net.FailRate + (0.00005 * (net.Layers - 1) * (net.Neurons - 1));
            if (_fcost > cost)
            {
                _fcost = cost;
                _minerrNN = net.TrainErrorMin;
                _maxerrNN = net.TrainErrorMax;
                _meanerrNN = net.TrainError;
                _miderrNN = net.TrainErrorMedian;
                _dverrNN = net.TrainErrorDev;
                return true;
            }
            return false;
        }
        private bool SameAs(NNGene g)
        {
            bool h = (_hidden == null) && (g._hidden == null);
            if ((_hidden != null) && (g._hidden != null) && (_hidden.Length == g._hidden.Length))
            {
                h = true;
                for (int nh = 0; nh < _hidden.Length; nh++)
                {
                    if (_hidden[nh] != g._hidden[nh])
                    {
                        h = false;
                        break;
                    }
                }
            }
            return (_alpha == g._alpha) && (_input == g._input) && h;
        }
        public bool Equals(NNGene g)
        {
            return _fcost == g._fcost;
        }
        public int CompareTo(NNGene g)
        {
            if (_fcost < g._fcost)
            {
                return -1;
            }
            if (_fcost > g._fcost)
            {
                return 1;
            }
            return 0;
        }
        public Network BuildNN(int inputs)
        {
            return new Network(inputs, _input, _hidden, _alpha);
        }
    }
}
